@@ -374,32 +374,60 @@ class AgentLoop:
             result: mcputil.Result = await func.call(
                 call_id=function_call.call_id, **args
             )
-            async for event in result.events():
-                if isinstance(event, mcputil.ProgressEvent):
-                    # Report progress to the context.
-                    tool_ctx.report_progress(
-                        progress=event.progress or 0,
-                        total=event.total or 0,
-                        message=event.message or "",
-                    )
-                elif isinstance(event, mcputil.OutputEvent):
-                    return ToolCallOutputItem(
-                        raw_item=ResponseFunctionToolCallOutputItem(
-                            id="",
-                            call_id=function_call.call_id,
-                            output=str(event.output),
-                            type="function_call_output",
-                            status="completed",
-                        ),
-                        output=event.output,
-                        type="tool_call_output_item",
-                    )
+            try:
+                async for event in result.events():
+                    if isinstance(event, mcputil.ProgressEvent):
+                        # Report progress to the context.
+                        tool_ctx.report_progress(
+                            progress=event.progress or 0,
+                            total=event.total or 0,
+                            message=event.message or "",
+                        )
+                    elif isinstance(event, mcputil.OutputEvent):
+                        return ToolCallOutputItem(
+                            raw_item=ResponseFunctionToolCallOutputItem(
+                                id="",
+                                call_id=function_call.call_id,
+                                output=str(event.output),
+                                type="function_call_output",
+                                status="completed",
+                            ),
+                            output=event.output,
+                            type="tool_call_output_item",
+                        )
+            except Exception as exc:
+                # Handle tool exceptions and report them as incomplete outputs.
+                return ToolCallOutputItem(
+                    raw_item=ResponseFunctionToolCallOutputItem(
+                        id="",
+                        call_id=function_call.call_id,
+                        output=str(exc),
+                        type="function_call_output",
+                        status="incomplete",
+                    ),
+                    output=str(exc),
+                    type="tool_call_output_item",
+                )
         else:
-            raw_result = func(**args)
-            if inspect.isawaitable(raw_result):
-                result = await raw_result
-            else:
-                result = raw_result
+            try:
+                raw_result = func(**args)
+                if inspect.isawaitable(raw_result):
+                    result = await raw_result
+                else:
+                    result = raw_result
+            except Exception as exc:
+                # Handle tool exceptions and report them as incomplete outputs.
+                return ToolCallOutputItem(
+                    raw_item=ResponseFunctionToolCallOutputItem(
+                        id="",
+                        call_id=function_call.call_id,
+                        output=str(exc),
+                        type="function_call_output",
+                        status="incomplete",
+                    ),
+                    output=str(exc),
+                    type="tool_call_output_item",
+                )
 
             return ToolCallOutputItem(
                 raw_item=ResponseFunctionToolCallOutputItem(
@@ -421,7 +449,11 @@ class AgentLoop:
         response_format: dict | None = None,
     ) -> AsyncIterator[ChatCompletionChunk]:
         context_variables = defaultdict(str, context_variables)
-        messages = Converter.items_to_messages(history)
+        try:
+            messages = Converter.items_to_messages(history)
+        except Exception as exc:
+            yield self.create_chunk(f"Failed to convert items to messages: {exc}")
+            return
 
         if self.agent.system:
             messages.insert(
@@ -432,7 +464,12 @@ class AgentLoop:
                 },
             )
 
-        tools = [function_to_jsonschema(tool) for tool in self.agent.tools]
+        try:
+            tools = [function_to_jsonschema(tool) for tool in self.agent.tools]
+        except Exception as exc:
+            yield self.create_chunk(f"Failed to convert tools to JSON schema: {exc}")
+            return
+
         # hide context_variables from model
         for tool in tools:
             params = tool["function"]["parameters"]
@@ -464,20 +501,22 @@ class AgentLoop:
         except Exception as exc:
             # Return the error in form of a completion chunk.
             model = self.agent.model.id
-            chunk = ChatCompletionChunk(
-                id=FAKE_ID,
-                choices=[
-                    Choice(
-                        delta=ChoiceDelta(
-                            role="assistant",
-                            content=f"Failed to chat with {model}: {exc}",
-                        ),
-                        finish_reason="stop",
-                        index=0,
-                    )
-                ],
-                created=int(time.time()),
-                model=model,
-                object="chat.completion.chunk",
-            )
-            yield chunk
+            yield self.create_chunk(f"Failed to chat with {model}: {exc}")
+
+    def create_chunk(self, content: str) -> ChatCompletionChunk:
+        return ChatCompletionChunk(
+            id=FAKE_ID,
+            choices=[
+                Choice(
+                    delta=ChoiceDelta(
+                        role="assistant",
+                        content=content,
+                    ),
+                    finish_reason="stop",
+                    index=0,
+                )
+            ],
+            created=int(time.time()),
+            model=self.agent.model.id,
+            object="chat.completion.chunk",
+        )
